/*
 * Copyright (c) 2008, 2017 Nicira, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef DHPARAMS_H
#define DHPARAMS_H 1

#include <inttypes.h>
#include <openssl/dh.h>

DH *get_dh1024(void);
DH *get_dh2048(void);
DH *get_dh4096(void);

#endif /* dhparams.h */
