/*
 * Copyright (c) 2011 Nicira, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <config.h>

#include "ofproto/ofproto.h"

#include "dpif.h"

/* This function is in a separate file because it is the only ofproto function
 * required by ovs-ofctl, which allows ovs-ofctl to avoid linking in extra
 * baggage. */

/* Parses 'ofproto_name', which is of the form [type@]name into component
 * pieces that are suitable for passing to ofproto_create().  The caller must
 * free 'name' and 'type'. */
void
ofproto_parse_name(const char *ofproto_name, char **name, char **type)
{
    dp_parse_name(ofproto_name, name, type);
}

